import logging
from dataclasses import dataclass
from typing import Optional

import requests
from mlflow.tracing.destination import TraceDestination

from databricks.rag_eval import context

_logger = logging.getLogger(__name__)


@dataclass
class DatabricksAgentMonitoringDestination(TraceDestination):
    """
    A destination representing a Databricks agent monitor.
    Sets databricks_monitor_id to the endpoint name for the monitor for the provided
    monitoring table, if a monitor is found. Sets to None otherwise.
    """

    databricks_monitor_id: Optional[str]

    @property
    def type(self) -> str:
        return "databricks-ingestion-endpoint"

    @context.eval_context
    def __init__(self, databricks_monitoring_table: str):
        managed_evals_client = context.get_context().build_managed_evals_client()

        try:
            monitor = managed_evals_client.get_monitor(
                monitoring_table=databricks_monitoring_table,
            )

            self.databricks_monitor_id = monitor.ingestion_endpoint_name
        except requests.HTTPError as e:
            if e.response.status_code == 404:
                _logger.warning(
                    f"Unable to find a monitor for monitoring table {databricks_monitoring_table}. "
                    f"Traces will not be available in Databricks until a monitor is created "
                    f"and this agent is restarted."
                )
                self.databricks_monitor_id = None
            else:
                raise
